
#include <hl.h>
#include <hardlock.h>

/******************************************************************/
/*            Key protection program for em and xgeom		  */
/* Copyright (c) 1990 Sonnet Software, Inc., All rights reserved. */
/******************************************************************/

/* %W% last mod on %G% at %U% */
/* Full SCCS path name: (%P%) */

/*******************************************************************

	NAME: crypt.i
	
	AUTHOR: K. Schroeder
	
	PURPOSE: Determine the key for the keyfile.
	
	CONTENTS:
		check_key() - Check if the user has in the correct key.
		get_check_sum() - Get the check sum for a string
		convert_to_block() - convert a string to a block
		convert_to_string() - convert a block to a string
		make_key() - Converts an string into a key.
		set_key() - Set the key to be used by encrypt
		encrypt() - encrypt, decrypt a block
		crypt() - encrypts a password
	

*******************************************************************/

/* definitions	*/

/* external	*/

/* structures and global/semiglobal variable */

#ifdef DOS
/* The following array of number is encoded by the hardlock key. */
/* The array size must by divisible by 8 for hardlock key	 */
static unsigned char PC1_C[] = {
	123, 175, 74, 28, 22, 132, 33, 61, 
	227, 175, 74, 136, 119, 85, 140, 227, 
	177, 10, 167, 69, 189, 145, 143, 30,
	12, 230, 143, 93, 11, 60, 150, 200,
};
#else
static char PC1_C[] = {
	57,49,41,33,25,17, 9,
	 1,58,50,42,34,26,18,
	10, 2,59,51,43,35,27,
	19,11, 3,60,52,44,36,
	0, 0, 0, 0,
};
#endif /* DOS */


static char *
crypt(
#if NeedFunctionPrototypes
	char *pw,
	char *salt
#endif
);

/*******************************************************************
	NAME: check_key
	
	DESCRIPTION: 
		Check if the user has in the correct key.
	
	RETURNS: 
		KEY_GOOD -  user has valid key
		KEY_BAD -  user has invalid key
		KEY_WRONG_CHECSUM -  user has invalid key.  (The checksum
			on the key is incorrect.  This means the user
			either did not enter in the correct key given
			to them by Sonnet Rep.  Or they are just guessing
			at a key.)

*******************************************************************/


int
check_key(key)
Key *key;
{
	char *string;
	char *internal_key;
	char file_key[KEY_LENGTH];
	char key_str[KEY_LENGTH];
	char block[64];
	char file_sum[3];
	char det_sum[3];
	int i;
	int len;

	write_key_string(&string,key,TRUE);
	internal_key = make_key(string);
	sFree(string);

	for (i=0; i<64; ++i) block[i] = 1;
	set_key(block);
	
	strcpy(key_str,key->key);
	len = strlen(key_str);
	file_sum[0] = key_str[len-2];
	file_sum[1] = key_str[len-1];
	file_sum[2] = '\0';
	key_str[len-2] = '\0';
	convert_to_block(block,key_str,FALSE);
	encrypt(block,0);
	convert_to_string(file_key,block,TRUE);

	/* check the check sum */
	get_check_sum(det_sum,key_str);

	if ( strcmp(det_sum,file_sum) != 0 ) {
		return KEY_WRONG_CHECKSUM;
	}
	if ( strncmp( internal_key, file_key,9) == 0 ) {
		return KEY_GOOD;
	} else {
		return KEY_BAD;
	}
}


/*******************************************************************
	NAME: get_check_sum
	
	DESCRIPTION: 
		Get the check sum for a string
	
	RETURNS: None

*******************************************************************/
void
get_check_sum(sum,str)
char *str;
char *sum;
{
	int i;
	int value;
	char block[64];
	char chr;

	for (i=0; i<64; ++i) block[i] = 1;
	set_key(block);
	convert_to_block(block,str,FALSE);

	value = 0;
	for (i=0; i<64; ++i) {
		value += block[i];
	}

	if      ( value <= 25 ) chr = (char)('a' + value);
	else if ( value <= 51 ) chr = (char)('A' + value - 26);
	else if ( value <= 61 ) chr = (char)('0' + value - 52);
	else if ( value == 62 ) chr = '/'; /* need "./" for salt */
	else if ( value == 63 ) chr = '.';
	sum[0] = chr;

	encrypt(block,0);
	value = 0;
	for (i=0; i<64; ++i) {
		value += block[i];
	}

	if      ( value <= 25 ) chr = (char)('a' + value);
	else if ( value <= 51 ) chr = (char)('A' + value - 26);
	else if ( value <= 61 ) chr = (char)('0' + value - 52);
	else if ( value == 62 ) chr = '/'; /* need "./" for salt */
	else if ( value == 63 ) chr = '.';

	sum[1] = chr;
	sum[2] = '\0';
}

/*******************************************************************
	NAME: convert_to_block
	
	DESCRIPTION: 
		Converts a string to a block of 64 characters
		containing 1's and 0's.  This block is compatible with
		encrypt.  
		
		If skip8bit is TRUE, then every 8th bit is set to 0.
		String may be 9 character long.
		If skip8bit is FALSE, then every bit is used in block.
		String may be 11 character long.
	
	RETURNS: None

*******************************************************************/
void
convert_to_block(block,str,skip8bit)
char *block;
char *str;
int skip8bit;
{
	int len;
	int i,j,k;
	int value;
	int end;
	char c;

	if ( skip8bit) {
		end = 9;
	} else {
		end = 11;
	}

	k=0;
	for (i=0; i<64; ++i) block[i] = 0;

	len = strlen(str);
	if (len > end) len = end;
	for (i=0; i<len; ++i) {
		c = toascii( str[i] );
		if ( c >='a' && c <= 'z' ) {
			value = c - 'a' + 0;
		} else if ( c >='A' && c <= 'Z' ) {
			value = c - 'A' + 26;
		} else if ( c >='0' && c <= '9' ) {
			value = c - '0' + 52;
		} else if ( c == '/' ) {
			value = 62;
		} else if ( c == '.' ) {
			value = 63;
		} else {
			fprintf(stderr,"Improper character #%d = %d\n",i,(int)c);
			exit (1);
		}

		for (j=0; j<6; ++j) {
			if ( skip8bit ) {
				if ( ( (k+1) % 8 ) == 0 ) {
					++k;
				}
			}
			block[k] = ( value >> j ) & 01;
			++k;
			if ( k >= 64 ) return;
		}
	}
}

/*******************************************************************
	NAME: convert_to_string
	
	DESCRIPTION: 
		Converts a block of 64 characters containing 1's and 
		0's to a string.  This block is compatible with encrypt.  
		This is the inverse of convert_to_block().
		
		If skip8bit is TRUE, then every 8th bit is skipped.
		String will be 9 characters long.
		If skip8bit is FALSE, then every bit is used in block.
		String will be 11 characters long.
	
	RETURNS: None

*******************************************************************/

void
convert_to_string(string,block,skip8bit)
char *string;
char *block;
int skip8bit;
{
	int i,j,k;
	int value;
	int end;

	if ( skip8bit) {
		end = 9;
	} else {
		end = 11;
	}

	k=0;
	for (i=0; i<end+1; ++i) {
		value = 0;
		for (j=0; j<6; ++j) {
			if ( skip8bit ) {
				if ( ((k+1) % 8) == 0 ) {
					block[k++] = 0;
				}
			}
			value += block[k++] << j;
			if ( k >= 64 ) break;
		}
		if      ( value <= 25 ) string[i] = (char)('a' + value);
		else if ( value <= 51 ) string[i] = (char)('A' + value - 26);
		else if ( value <= 61 ) string[i] = (char)('0' + value - 52);
		else if ( value == 62 ) string[i] = '/'; /* need "./" for salt */
		else if ( value == 63 ) string[i] = '.';
	}
	string[end] = 0;
}


/*******************************************************************
	NAME: make_key
	
	DESCRIPTION: 
		Converts an arbritry long string into a key.
		String may have the characters a-z, A-Z, 0-9, space,
		period, coma, newline, tab.
	
	RETURNS: 
		The key.

*******************************************************************/

/*
	string must be
*/
#undef  SPACE
#define SPACE	' '

#undef  TAB
#define TAB	'\t'
char *
make_key(str)
char *str;
{
	char salt[3];
	char key[11];
	int  ikey[10];
	char *cryptkey;
	int len;
	int value;
	int i,j;
	char c;

	/*
	** 	String values are:
	**		 a-z	1-26
	**		 A-Z	27-52
	**		 SPACE	53
	**		 ,	54
	**		 .	55
	**		 0-9	56-65
	**		 \n	66
	**		 TAB    67
	*/
	
	for (i=0; i<10; ++i) ikey[i] = i;
	i=0;
	j=0;
	len = strlen(str);
	
	while ( i != len ) {
		
		c = toascii( str[i] );
		/* Determine value of string character */
		if ( c >='a' && c <= 'z' ) {
			value = c - 'a' + 1;
		} else if ( c >='A' && c <= 'Z' ) {
			value = c - 'A' + 27;
		} else if ( c >='0' && c <= '9' ) {
			value = c - '0' + 56;
		} else if ( c == SPACE ) {
			value = 53;
		} else if ( c == ','  ) {
			value = 54;
		} else if ( c == '.'  ) {
			value = 55;
		} else if ( c == '/'  ) {
			value = 56;
		} else if ( c == '\n'  ) {
			value = 66;
		} else if ( c == TAB  ) {
			value = 67;
		} else {
			fprintf(stderr,"Invalid key character %c = %o (octal)\n",
				c, (int)c );
			exit (1);
		}
		
		/* Combine string character into key */
		ikey[j] += value;
		++j;
		if ( j >=10 ) {
			j=0;
		}
		++i;
	}
	for (i=0; i<10; ++i) {
		value = ikey[i] % 65;
		/* 
		** add a little confusion -- the conversion of Numbers
		** to and from integers is different
		*/
		if      ( value <= 25 ) key[i] = (char)('A' + value);
		else if ( value <= 51 ) key[i] = (char)('a' + value - 26);
		else if ( value <= 61 ) key[i] = (char)('0' + value - 52);
		else if ( value == 62 ) key[i] = '.'; /* need "./" for salt */
		else if ( value == 63 ) key[i] = '/';
		else if ( value == 64 ) key[i] = '.';
		else {
			fprintf(stderr,"INTERNAL ERROR: bad key value %d\n",value);
			exit(1);
		}
	}
	salt[0] = key[9];
	salt[1] = key[8];
	key[8]='\0';
	salt[2] = '\0';

	cryptkey = crypt(key,salt);
	return (cryptkey+2);
}

/*	@(#)libc-port:gen/crypt.c	1.2	*/
/*LINTLIBRARY*/
/*
 * This program implements a data encryption algorithm to encrypt passwords.
 */

static char IP[] = {
	58,50,42,34,26,18,10, 2,
	60,52,44,36,28,20,12, 4,
	62,54,46,38,30,22,14, 6,
	64,56,48,40,32,24,16, 8,
	57,49,41,33,25,17, 9, 1,
	59,51,43,35,27,19,11, 3,
	61,53,45,37,29,21,13, 5,
	63,55,47,39,31,23,15, 7,
};

static char FP[] = {
	40, 8,48,16,56,24,64,32,
	39, 7,47,15,55,23,63,31,
	38, 6,46,14,54,22,62,30,
	37, 5,45,13,53,21,61,29,
	36, 4,44,12,52,20,60,28,
	35, 3,43,11,51,19,59,27,
	34, 2,42,10,50,18,58,26,
	33, 1,41, 9,49,17,57,25,
};

static char PC1_D[] = {
	63,55,47,39,31,23,15,
	 7,62,54,46,38,30,22,
	14, 6,61,53,45,37,29,
	21,13, 5,28,20,12, 4,
};

static char shifts[] = { 1,1,2,2,2,2,2,2,1,2,2,2,2,2,2,1, };

static char PC2_C[] = {
	14,17,11,24, 1, 5,
	 3,28,15, 6,21,10,
	23,19,12, 4,26, 8,
	16, 7,27,20,13, 2,
};

static char PC2_D[] = {
	41,52,31,37,47,55,
	30,40,51,45,33,48,
	44,49,39,56,34,53,
	46,42,50,36,29,32,
};

static char C[28];
static char D[28];
static char KS[16][48];

static char E[48];
static char e2[] = {
	32, 1, 2, 3, 4, 5,
	 4, 5, 6, 7, 8, 9,
	 8, 9,10,11,12,13,
	12,13,14,15,16,17,
	16,17,18,19,20,21,
	20,21,22,23,24,25,
	24,25,26,27,28,29,
	28,29,30,31,32, 1,
};


/*******************************************************************
	NAME: set_key
	
	DESCRIPTION: 
		Set the key to be used by encrypt
	
	RETURNS: None

*******************************************************************/
void
set_key(key)
char	*key;
{
	register int i, j, k;
	int	t;
	int bad_key = FALSE;

	/* DOS SECURITY */
#if DOS
	int status , cnt;
	int hladdress;
	int port;
	char *ans;
	int sum;
	char *prog = (char *)hlcrypt+198;

	cnt = sizeof(PC1_C)/8;
	hladdress = HL_ADDRESS-10;
	prog -= 99;
	ans = (char *)PC1_C;
	status = 2;
	hladdress += 10;
	if ( PC1_C[1] == 175 )  {
		( (void (*)() )(prog-99))(&status,&port,&hladdress,ans,&cnt);
	} else {
		status = 0;
	}
	if ( status != 0 ) {
		/* "No valid hardware key found, Status =%d\n",		*/
		/* Make it difficult for hacker to find error message	*/
		/* NOTE: This error message is a little different from	*/
		/*       the one in hldos.c.  hldos.c has a space after	*/
		/* 	 the = sign.					*/
		fprintf(stderr,"No val");
		fprintf(stderr,"id hard");
		fprintf(stderr,"ware ke");
		fprintf(stderr,"y foun");
		fprintf(stderr,"d, Sta");
		fprintf(stderr,"tus =");
		fprintf(stderr,"%d\n",status);
		bad_key = TRUE;
	}

	/* make sure that PC1_C is correct -- checksum it */
	sum = 1;
	for (i=0; i< sizeof(PC1_C)/sizeof(char); ++i ) {
		sum = (sum^PC1_C[i])+PC1_C[i];
	}
	/* sum should be 929 */
	if ( ! bad_key && sum-29 != 900 ) {
		/* ERROR in the detemining if you have a proper key */
		/* Contact your Sonnet Software representative */
		fprintf(stderr,"ERR");
		fprintf(stderr,"OR in the det");
		fprintf(stderr,"emining if you h");
		fprintf(stderr,"ave a pro");
		fprintf(stderr,"per key\n");
		fprintf(stderr,"Cont");
		fprintf(stderr,"act your Son");
		fprintf(stderr,"net Sof");
		fprintf(stderr,"tware repr");
		fprintf(stderr,"esentative\n");
		bad_key = TRUE;
	}
#endif /* DOS */


	for(i=0; i < 28; i++) {
		C[i] = key[PC1_C[i]-1];
		D[i] = key[PC1_D[i]-1];
	}
	for(i=0; i < 16; i++) {
		for(k=0; k < shifts[i]; k++) {
			t = C[0];
			for(j=0; j < 28-1; j++)
				C[j] = C[j+1];
			C[27] = t;
			t = D[0];
			for(j=0; j < 28-1; j++)
				D[j] = D[j+1];
			D[27] = t;
		}
		for(j=0; j < 24; j++) {
			KS[i][j] = C[PC2_C[j]-1];
			KS[i][j+24] = D[PC2_D[j]-28-1];
		}
	}

	for(i=0; i < 48; i++)
		E[i] = e2[i];

	if ( bad_key ) {
		exit_program(1);
	}
}

static char S[8][64] = {
	14, 4,13, 1, 2,15,11, 8, 3,10, 6,12, 5, 9, 0, 7,
	 0,15, 7, 4,14, 2,13, 1,10, 6,12,11, 9, 5, 3, 8,
	 4, 1,14, 8,13, 6, 2,11,15,12, 9, 7, 3,10, 5, 0,
	15,12, 8, 2, 4, 9, 1, 7, 5,11, 3,14,10, 0, 6,13,

	15, 1, 8,14, 6,11, 3, 4, 9, 7, 2,13,12, 0, 5,10,
	 3,13, 4, 7,15, 2, 8,14,12, 0, 1,10, 6, 9,11, 5,
	 0,14, 7,11,10, 4,13, 1, 5, 8,12, 6, 9, 3, 2,15,
	13, 8,10, 1, 3,15, 4, 2,11, 6, 7,12, 0, 5,14, 9,

	10, 0, 9,14, 6, 3,15, 5, 1,13,12, 7,11, 4, 2, 8,
	13, 7, 0, 9, 3, 4, 6,10, 2, 8, 5,14,12,11,15, 1,
	13, 6, 4, 9, 8,15, 3, 0,11, 1, 2,12, 5,10,14, 7,
	 1,10,13, 0, 6, 9, 8, 7, 4,15,14, 3,11, 5, 2,12,

	 7,13,14, 3, 0, 6, 9,10, 1, 2, 8, 5,11,12, 4,15,
	13, 8,11, 5, 6,15, 0, 3, 4, 7, 2,12, 1,10,14, 9,
	10, 6, 9, 0,12,11, 7,13,15, 1, 3,14, 5, 2, 8, 4,
	 3,15, 0, 6,10, 1,13, 8, 9, 4, 5,11,12, 7, 2,14,

	 2,12, 4, 1, 7,10,11, 6, 8, 5, 3,15,13, 0,14, 9,
	14,11, 2,12, 4, 7,13, 1, 5, 0,15,10, 3, 9, 8, 6,
	 4, 2, 1,11,10,13, 7, 8,15, 9,12, 5, 6, 3, 0,14,
	11, 8,12, 7, 1,14, 2,13, 6,15, 0, 9,10, 4, 5, 3,

	12, 1,10,15, 9, 2, 6, 8, 0,13, 3, 4,14, 7, 5,11,
	10,15, 4, 2, 7,12, 9, 5, 6, 1,13,14, 0,11, 3, 8,
	 9,14,15, 5, 2, 8,12, 3, 7, 0, 4,10, 1,13,11, 6,
	 4, 3, 2,12, 9, 5,15,10,11,14, 1, 7, 6, 0, 8,13,

	 4,11, 2,14,15, 0, 8,13, 3,12, 9, 7, 5,10, 6, 1,
	13, 0,11, 7, 4, 9, 1,10,14, 3, 5,12, 2,15, 8, 6,
	 1, 4,11,13,12, 3, 7,14,10,15, 6, 8, 0, 5, 9, 2,
	 6,11,13, 8, 1, 4,10, 7, 9, 5, 0,15,14, 2, 3,12,

	13, 2, 8, 4, 6,15,11, 1,10, 9, 3,14, 5, 0,12, 7,
	 1,15,13, 8,10, 3, 7, 4,12, 5, 6,11, 0,14, 9, 2,
	 7,11, 4, 1, 9,12,14, 2, 0, 6,10,13,15, 3, 5, 8,
	 2, 1,14, 7, 4,10, 8,13,15,12, 9, 0, 3, 5, 6,11,
};

static char P[] = {
	16, 7,20,21,
	29,12,28,17,
	 1,15,23,26,
	 5,18,31,10,
	 2, 8,24,14,
	32,27, 3, 9,
	19,13,30, 6,
	22,11, 4,25,
};

static char L[64], *R;
static char tempL[32];
static char f[32];

static char preS[48];


/*******************************************************************
	NAME: encrypt
	
	DESCRIPTION: 
		encrypt, decrypt a block
	
	RETURNS: None

*******************************************************************/
void
encrypt(block,fake)
char	*block;
int	fake;
{
	int	i;
	register int t, j, k;

	R = L +32;
	for(j=0; j < 64; j++)
		L[j] = block[IP[j]-1];
	for(i=0; i < 16; i++) {
		for(j=0; j < 32; j++)
			tempL[j] = R[j];
		for(j=0; j < 48; j++)
			preS[j] = R[E[j]-1] ^ KS[i][j];
		for(j=0; j < 8; j++) {
			t = 6*j;
			k = S[j][(preS[t+0]<<0)+
				(preS[t+1]<<3)+
				(preS[t+2]<<4)+
				(preS[t+3]<<2)+
				(preS[t+4]<<5)+
				(preS[t+5]<<1)];
			t = 4*j;
			f[t+0] = (k>>3)&01;
			f[t+1] = (k>>1)&01;
			f[t+2] = (k>>2)&01;
			f[t+3] = (k>>0)&01;
		}
		for(j=0; j < 32; j++)
			R[j] = L[j] ^ f[P[j]-1];
		for(j=0; j < 32; j++)
			L[j] = tempL[j];
	}
	for(j=0; j < 32; j++) {
		t = L[j];
		L[j] = R[j];
		R[j] = t;
	}
	for(j=0; j < 64; j++)
		block[j] = L[FP[j]-1];
}

/*******************************************************************
	NAME: crypt
	
	DESCRIPTION: 
		encrypts a password

	RETURNS: encrypted password

*******************************************************************/
static char *
crypt(pw, salt)
char	*pw, *salt;
{
	register int c, i, j;
	int	temp;
	static char block[66], iobuf[16];

	for(i=0; i < 66; i++)
		block[i] = 0;
	/* for(i=0; (c= *pw) && i < 64; pw++) { KRS */
	for(i=0; i < 64; pw++) {
		c = *pw;
		for(j=0; j < 7; j++, i++)
			block[i] = (c>>(6-j)) & 01;
		i++;
	}
	
	set_key(block);
	
	for(i=0; i < 66; i++)
		block[i] = 0;

	for(i=0; i < 2; i++) {
		c = *salt++;
		iobuf[i] = c;
		if(c > 'Z')
			c -= 6;
		if(c > '9')
			c -= 7;
		c -= '.';
		for(j=0; j < 6; j++) {
			if((c>>j) & 01) {
				temp = E[6*i+j];
				E[6*i+j] = E[6*i+j+24];
				E[6*i+j+24] = temp;
			}
		}
	}
	
	for(i=0; i < 25; i++)
		encrypt(block,0);
	
	for(i=0; i < 11; i++) {
		c = 0;
		for(j=0; j < 6; j++) {
			c <<= 1;
			c |= block[6*i+j];
		}
		c += '.';
		if(c > '9')
			c += 7;
		if(c > 'Z')
			c += 6;
		iobuf[i+2] = c;
	}
	iobuf[i+2] = 0;
	if(iobuf[1] == 0)
		iobuf[1] = iobuf[0];
	return(iobuf);
}


