
/*Utility routines for calculating radiation patterns from currents */

#include "local.h"
#include <stdio.h> 
#include <math.h>
#include <ctype.h>
#include "son.h"
#include "em.typ"
#include "jxy.typ"
#include "complex2.h"
#include "patdata.h"

extern char prog_name[];

/*
 *  double sinc(double x)
 *    Compute sinc function
 *
 */
static double sinc(x)
     double x;
{
  if (fabs(x) < EPS) return (1.0);
  else               return (sin(x)/x);
}

/* 
 *  void Cpolar2rect(complex *c1, *c2);
 * Covert a complex from polar form to rect form                            
 * 
 */
static void Cpolar2rect(c1,c2) 
     complex *c1,*c2;
{
  c2->Re=c1->Re*cos(c1->Im);
  c2->Im=c1->Re*sin(c1->Im);
} 

/*
 * void Crect2polar(complex *c1,*c2)
 *    Covert a complex from rect form to polar form               
 *
 */
static void Crect2polar(c1,c2) 
     complex *c1,*c2;
{
  c2->Re=hypot(c1->Re,c1->Im);
  if (c2->Re < TINY) c2->Im=0.0;
  else               c2->Im=atan2(c1->Re,c1->Im);
} 

/*
 *  void Cadd(complex *c1,*c2,*c)
 *      Add two complex numbers 
 *
 */
static void Cadd(c1,c2,c) 
     complex *c1,*c2,*c;
{
  c->Re=c1->Re+c2->Re;
  c->Im=c1->Im+c2->Im;
} 

/*
 *  void Cexpj(double phase, complex *c)
 *    Compute exp(j*phase)
 *
 */
static void   Cexpj(tt,c) 
     Rtyp tt; 
     complex *c;
{
  c->Re=cos(tt);
  c->Im=sin(tt);
} 

/*
 *  void WpatData_free(WPatData *ad)
 *    Free up memory pointed by WPatData
 *
 */
void WPatData_free(ad)
     WPatData *ad;
{ 
  if (ad != NULL) {
    if(ad->freq  != NULL)  { sFree((char *)ad->freq);  ad->freq=NULL;}
    if(ad->theta != NULL)  { sFree((char *)ad->theta); ad->theta=NULL;}
    if(ad->phi   != NULL)  { sFree((char *)ad->phi);   ad->phi=NULL;}
    if(ad->field != NULL)  { sFree((char *)ad->field); ad->field=NULL;}
  }
}

/*
 * void WPatData_init(WPatGenTmp *td, WPatData *ad)
 *   Initialize WPatData
 *
 */
void WPatData_init(td,ad)
     WPatGenTmp *td;
     WPatData *ad;
{ 
  Jdat *jd;
  int i,ifreq,nv;
  int *findx;
  double *vf;
  WPatRange *pf;
  extern float *Range2FVector(/* WPatRange *pf, int *nv */);
  extern double *Range2DVector(/* WPatRange *pf, int *nv */);

  jd=td->jd;
  /* check ground plane */
  if (fabs(jd->box->Bot.rdc) < EPS)  /* rc=0 */
    td->iground=YES;
  else  if (fabs(jd->box->Bot.rdc) > EPS && jd->box->Bot.rdc < 300.0) {
    /* 0 < rc < 300 */
    printf("Warning: the ground plane is not perfectly conductor!\n");
    td->iground=YES;
  }
  else if (fabs(jd->box->Bot.rdc-377) > 1.0)  {
    printf("Warning: the ground plane is assumed non-existing!\n");
    td->iground=NO;
  }
  else 
    td->iground=NO;
  
  /* check top cover */
  if (fabs(jd->box->Top.rdc-377.0) > 1.0) 
    printf("Warning: the top cover is not set to 377 ohms/m/m!\n");
  
  ad->ipatvu=td->icoord;
  
  /* set up the theta range */
  ad->theta=Range2FVector(&(td->ang1),&(ad->ntheta));

  /* set up the phi range */
  ad->phi=Range2FVector(&(td->ang2),&(ad->nphi));

  /* set up the freq range */
  pf=td->freq;
  findx=(int *)new(sizeof(int)*jd->nfreq);
  for (i=0; i < jd->nfreq; i++) findx[i]=NO;
  while (pf != NULL) {
    if (Bit1(pf->unit) == FREQ_GHZ) {
      pf->start *= 1000.0;
      pf->stop  *= 1000.0;
      pf->step  *= 1000.0;
      pf->unit  = 0;
    }
    vf=Range2DVector(pf,&nv);
    PosSubVecInVec(vf,nv,jd->freq,jd->nfreq,findx);
    pf=pf->next;
    sFree((char *)vf);
  }
  ifreq=0;
  for (i=0; i < jd->nfreq; i++) 
    if (findx[i] != NO) ifreq++;

  ad->nfreq=ifreq;
  ad->freq=(float *)new(sizeof(float)*ad->nfreq);
  td->freq_indx=(int *)new(sizeof(float)*ad->nfreq);

  ifreq=0;
  for (i=0; i < jd->nfreq; i++) {
    if (findx[i] != NO) {
      ad->freq[ifreq]=jd->freq[i];
      td->freq_indx[ifreq]=findx[i];
      ifreq++;
    }
  }
  sFree((char *)findx);
  
  /* set up para */
  ad->npara=1;
  ad->para=(float *)new(sizeof(float)*ad->npara);
  for (i=0; i < ad->npara; i++) *(ad->para+i)=0.0;
  td->npt=ad->ntheta*ad->nphi;
  for (i=0; i < ad->npara; i++) *(ad->para+i) = 0.0;
  ad->nfield=td->npt*ad->nfreq*4;
  ad->field=(float *)new(sizeof(float)*ad->nfield);
  for (i=0; i < ad->nfield; i++) *(ad->field+i)=0.0;

  /* misc */
  td->nbyte=(int) ((td->npt*4*sizeof(float)+
		    td->jd->box->mat_size*8*sizeof(float))*1e-6) + 1;
  strcpy(ad->FreqUnit,"MHz");
}

/*
 * void WPatData_write_header(WPatGenOpt *opt, WPatGenTmp *td, WPatData *ad)
 *  Write the header to file
 *
 */
void WPatData_write_header(opt,td,ad)
     WPatGenOpt *opt;
     WPatGenTmp *td;
     WPatData *ad;
{
  int i;
  
  fprintf(td->patfptr,"PAT **** New Antenna Pattern Section ****\n");
  fprintf(td->patfptr,"PVR %s\n",td->version);
  fprintf(td->patfptr,"NAM %s\n",opt->jxyfile);
  fprintf(td->patfptr,"CMD ");
  for (i=0; i < opt->argc; i++) 
    fprintf(td->patfptr,"%s ",opt->argv[i]);
  fprintf(td->patfptr,"\nTIM %s",td->time_start);    
  fprintf(td->patfptr,"LIC %s\n",td->license_id);    
  fprintf(td->patfptr,"HST %s\n",td->host_id);    
  for (i=0; i < td->jd->box->num_ann; i++) 
    fprintf(td->patfptr,"ANN %s\n",*(td->jd->box->ann+i));    
  fprintf(td->patfptr,"GTM %s\n",td->jd->box->xgeomdate);      
  fprintf(td->patfptr,"JTM %s\n",td->jd->tim);      
  fprintf(td->patfptr,"PAT Antenna Control Analysis File Follows\n");
  fseek(td->confptr,0L,0);
  WFile_copy(td->confptr,td->patfptr);
  fprintf(td->patfptr,"END # end of file %s\n",opt->confile);
  fprintf(td->patfptr,"ALST1 %d # List of ANG1\n",ad->ntheta);
  for (i=0; i < ad->ntheta; i++) {
    fprintf(td->patfptr,"%13.6e ",*(ad->theta+i));
    if ((i+1)%4 == 0) fprintf(td->patfptr,"\n");
  }
  if (i%4 != 0) fprintf(td->patfptr,"\n");  
  fprintf(td->patfptr,"END\n");
  fprintf(td->patfptr,"ALST2 %d # List of ANG2\n",ad->nphi);
  for (i=0; i < ad->nphi; i++) {
    fprintf(td->patfptr,"%13.6e ",*(ad->phi+i));
    if ((i+1)%4 == 0) fprintf(td->patfptr,"\n");
  }
  if (i%4 != 0) fprintf(td->patfptr,"\n");  
  fprintf(td->patfptr,"END\n");
  fprintf(td->patfptr,"FLST %d # List of FREQ\n",ad->nfreq);
  for (i=0; i < ad->nfreq; i++) {
    fprintf(td->patfptr,"%13.6e ",*(ad->freq+i));
    if ((i+1)%4 == 0 && i != 0) fprintf(td->patfptr,"\n");
  }
  if (i%4 != 0) fprintf(td->patfptr,"\n");  
  fprintf(td->patfptr,"END\n");
  fprintf(td->patfptr,"PVU %d # pattern viewing coordinates\n",ad->ipatvu);
  fclose(td->confptr);
  fclose(td->patfptr);
}

/*
 *  void WPatData_write_data
          (int ifreq, WPatGenOpt *opt, WPatGenTmp *td, WPatData *ad)
 *    Write data per freq to file
 *
 */
void WPatData_write_data(ifreq,opt,td,ad)
     int ifreq;
     WPatGenOpt *opt;
     WPatGenTmp *td;
     WPatData *ad;
{
  int i;
  extern FILE *WFile_open(/* char *file, char *mode */);

  td->patfptr=fopen(opt->patfile,"r");
  if (td->patfptr == NULL) {
    fprintf(stderr,"Output data file %s has been moved!\n",opt->patfile);
    exit(1);
  }
  fclose(td->patfptr);
  td->patfptr=WFile_open(opt->patfile,"a");
  fprintf(td->patfptr,"USR 0:25 user, 0:00 system, 0:37 real\n");
  fprintf(td->patfptr,"FRE %13.6e # frequency in MHz\n",*(ad->freq+ifreq));
  fprintf(td->patfptr,"PDT\n");
  for (i=0; i < ad->npara; i++) {
    fprintf(td->patfptr,"%13.6e ",*(ad->para+i));
    if ( i != 0 && i%4 == 0) fprintf(td->patfptr,"\n");
  }
  if ( (i+1)%4 != 0) fprintf(td->patfptr,"\n");
  for (i=0; i < td->npt; i++) 
    fprintf(td->patfptr,"%13.6e %13.6e %13.6e %13.6e\n",
	    ad->field[4*i],ad->field[4*i+1],ad->field[4*i+2],ad->field[4*i+3]);
  WTime_local(td->time_cur);
  fprintf(td->patfptr,"END # current time: %s\n",td->time_cur);    
  fclose(td->patfptr);
}

/*
 * void WPatData_fill(WPatGenOpt *opt, WPatGenTmp *td, WPatData *ad)
 *   Fill WPatData
 *
 */
void WPatData_fill(opt,td,ad)
     WPatGenOpt *opt;
     WPatGenTmp *td;
     WPatData *ad;
{
  Pdat **pd0;
  Pdat *pd;
  Jdat *jd;
  int ifreq,f_indx,itheta,iphi,ilev;
  int nlevels,npt; /* Number of dilectric layers */
  float *pfield;
  Rtyp freq,t1,t2,kh;
  Rtyp kfact,xfact,yfact,kz; /*geometry normalization factor*/
  complex etheta,ephi,c0,c1,c2,c3,c4;
  Rtyp xy[4]; 
  Angles angle;
  LevelParameters *lpara,*ptrl;
  sTime *stime;
  extern Pdat *setup_pdat();
  static int first_warning=0;

  /* write the header first */
  WPatData_write_header(opt,td,ad);

  /* change the angle units to radians */
  for (itheta=0; itheta < ad->ntheta; itheta++)
    if (td->icoord != PTN_COORD_ARRAY) 
      *(ad->theta+itheta) *= DEG_2_RAD;
  for (iphi=0; iphi < ad->nphi; iphi++)
    if (td->icoord != PTN_COORD_ARRAY) 
      *(ad->phi+iphi) *= DEG_2_RAD;

  jd=td->jd;
  /*calculate pd for temporary use*/
  xy[0]=0.0; xy[1]=0.0; xy[2]=jd->box->Capm; xy[3]=jd->box->Capn;
  nlevels=jd->box->Num_levels;
  pd0=(Pdat **)new(sizeof(Pdat *)*nlevels);
  for (ilev=0; ilev <nlevels; ilev++) {
    pd=*(pd0+ilev)=setup_pdat(jd,ilev,xy,1.0,1.0);
    if (pd != NULL) 
      fill_pdat(pd,NULL,jd,CALC_JX|CALC_JY|CALC_EZ,STEP_F,0,0,1,NULL,0);
  }
  
  pd= *pd0;
  if (opt->print_j & 1) {
    printf("   Pixel     Jx -- Mag        Jx -- Phase \n");
    while (pd) {
      for (itheta=0; itheta < pd->ny; itheta++) {
	for (iphi=0; iphi < pd->nx; iphi++) {
	  ifreq=iphi+itheta*pd->nx;
	  printf("(%3d,%3d) %15g %10.1f\n",iphi,itheta,(pd->jx_mg)[ifreq],RAD_2_DEG*((pd->jx_ph)[ifreq]));
	}
      } 
      pd=pd->next;
    }
  }
  pd= *pd0;
  if ((opt->print_j >> 1) & 1) {
    printf("   Pixel     Jy -- Mag        Jy -- Phase \n");
    while (pd) {
      for (itheta=0; itheta < pd->ny; itheta++) {
	for (iphi=0; iphi < pd->nx; iphi++) {
	  ifreq=iphi+itheta*pd->nx;
	  printf("(%3d,%3d) %15g %10.1f\n",iphi,itheta,(pd->jy_mg)[ifreq],RAD_2_DEG*((pd->jy_ph)[ifreq]));
	}
      } 
      pd=pd->next;
    }
  }

  pd= *pd0;
  if ((opt->print_j >> 2) & 1) {
    printf("   Pixel     Ez -- Mag        Ez -- Phase \n");
    while (pd) {
      for (itheta=0; itheta < pd->ny; itheta++) {
	for (iphi=0; iphi < pd->nx; iphi++) {
	  ifreq=iphi+itheta*pd->nx;
	  printf("(%3d,%3d) %15g %10.1f\n",iphi,itheta,(pd->ez_mg)[ifreq],RAD_2_DEG*((pd->ez_ph)[ifreq]));
	}
      } 
      pd=pd->next;
    }
  }

  angle.coord=ad->ipatvu;
  lpara=(LevelParameters *)new(nlevels*sizeof(LevelParameters));
  c0.Re=0.0; c0.Im=0.0;
  stime=GetTime((sTime *)NULL);
  for (ifreq=0; ifreq < ad->nfreq; ifreq++) {
    pfield=ad->field;
    freq=*(ad->freq+ifreq);
    if (opt->verbose == YES) {
      if (ifreq == 1) {
	stime=GetTime(stime);
	printf("   Analysis time per frequency (mm:ss) -- %s\n",TimeToString(stime));
      }
      printf("Analyzing %g GHz\n",freq*0.001);
    }
    f_indx= *(td->freq_indx+ifreq);
    kfact=TWO_PI*freq*1.0e+06/VEL_LIGHT;
    xfact=kfact*jd->box->A/jd->box->Capm;
    yfact=kfact*jd->box->B/jd->box->Capn;
    /* calc level parameters */
    calc_level_parameters(kfact,jd,lpara);
    /* change current format from polar to rect */ 
    for (ilev=0; ilev < nlevels; ilev++) 
      change_jxy_2_rect(*(pd0+ilev),f_indx);
    /*Generate the plot data*/
    for (iphi=0; iphi<ad->nphi; iphi++) {
      for (itheta=0; itheta<ad->ntheta; itheta++) {
	angle.phi= *(ad->phi+iphi);
	angle.theta= *(ad->theta+itheta);
	etheta=c0; ephi=c0;
	if (change_ptn_coord(&angle,td->iground) != NO) {
	/*Calculate radiation from each layer of metals*/
	  calc_apprx_factor(&angle,lpara,nlevels,td->apprx_level,td->iground);
	  for (ilev=0; 
	       ilev < nlevels; 
	       ilev++) {
	    ptrl=lpara+ilev;
	    if (ptrl->total_reflection == YES) {
	      if (first_warning) {
		printf("Warning: theta=%.1g, phi=%.1g, total reflection occurs \n",
		       angle.theta*RAD_2_DEG, angle.phi*RAD_2_DEG);
		first_warning=0;
	      }
	      break;
	    }
	    calc_ptrn(&etheta,&ephi,
		      &angle,*(pd0+ilev),jd,f_indx,xfact,yfact,ptrl);
	  }  /* for ilev */
	}   /* change_ptn_coord */
	/*Convert from rect form to polar form*/
	Crect2polar(&etheta,&c1);
	*pfield++     = c1.Re; 
	*pfield++ = c1.Im;
	Crect2polar(&ephi,&c2);
	*pfield++ = c2.Re; 
	*pfield++ = c2.Im;
      } /* theta */
    } /* phi */
    /* calc paras */
    calc_input_power(jd,f_indx,ad->para); 
    /* output */
    WPatData_write_data(ifreq,opt,td,ad);
  }   /* freq */
  
  if (opt->verbose == YES && ad->nfreq == 1) {
    stime=GetTime(stime);
    printf("   Analysis time per frequency (mm:ss) -- %s\n",TimeToString(stime));
    printf("Patgen: done!\n");
  }

  /* free up all tmp memory */
  for (ilev=0; ilev < nlevels; ilev++) clear_pdat(*(pd0+ilev));
  sFree((char *)pd0);
  sFree((char *)lpara);
}

/*
 * Calculate radiation pattern from each layer 
 *
 */
calc_ptrn(etheta,ephi,angle,pd0,jd,istep,xfact,yfact,lpara)  
     complex *etheta,*ephi;
     Angles *angle;/*radaition direction*/
     Pdat *pd0;
     Jdat *jd;
     int istep;
     Rtyp xfact,yfact;/*geometry normalization factors*/ 
     LevelParameters *lpara;
{ 
  int ix,iy,dat_head,ipos;
  float *dmxp,*dpxp,*dmyp,*dpyp;
  Rtyp  dx,dy,xi,yi,x0,y0;
  Rtyp  t00,t01,t10,t11,tt1,tt2,ts;
  complex ct,ct1,ct2,ct3;
  complex Jxi,Jyi;
  complex field[2];
  Pdat *pd;

  /*Currents in pd is stored as  [iy][ix]
    where iy=0, ..., pd->ny; ix=0, ..., pd->nx; */

  ct.Re=0.0; ct.Im=0.0;
  for (ix=0; ix < 2; ix ++) field[ix]=ct;
  pd=pd0;
  while(pd) {
    dat_head = pd->nx*pd->ny*istep;
    dx=pd->dx*xfact;  x0=pd->x0*xfact;
    dy=pd->dy*yfact;  y0=pd->y0*yfact;
    ts= dx*dy;
    t00=ts*angle->cos_p;  t01= ts*angle->sin_p;
    t10=-ts*angle->sin_p; t11= ts*angle->cos_p;
    dmxp=pd->jx_mg+dat_head; dpxp=pd->jx_ph+dat_head;
    dmyp=pd->jy_mg+dat_head; dpyp=pd->jy_ph+dat_head;
    for (iy=0; iy<pd->ny; iy++) {
      yi=jd->box->Capn*yfact-y0-iy*dy;  /* change coord origin */
      if (jd->box->Xsym == 0)  {
	tt1=yi*angle->sin_p; 
	tt2=1.0;
      }
      else {
	tt1=jd->box->Capn*yfact*angle->sin_p;
	tt2=2.0*cos((jd->box->Capn*yfact*0.5-yi)*angle->sin_t*angle->sin_p);
      }
      for (ix=0; ix<pd->nx; ix++) {
	xi=x0+ix*dx; 
	/*Phase delay due to pixel location*/
	ct1.Im=(tt1+xi*angle->cos_p)*angle->sin_t; 
	ct1.Re=tt2;
	Cpolar2rect(&ct1,&ct);
	Jxi.Re= *dmxp++; Jxi.Im= *dpxp++;
	Jyi.Re= *dmyp++; Jyi.Im= *dpyp++; 	
	Cmult(&ct,&Jxi,&ct1);
	Cmult(&ct,&Jyi,&ct2);
	Cadd(field,    &ct1,field);
	Cadd(field+1,  &ct2,field+1);
      }/*iy*/
    }/*ix*/
    pd=pd->next;
  }/*while*/
  
  ct1.Re=t00*field[0].Re+t01*field[1].Re;
  ct1.Im=t00*field[0].Im+t01*field[1].Im;
  ct2.Re=t10*field[0].Re+t11*field[1].Re;
  ct2.Im=t10*field[0].Im+t11*field[1].Im;
  
  Cmult(&ct1,lpara->factor,field);
  Cadd(etheta,field,etheta);
  Cmult(&ct2,lpara->factor+1,field+1);
  Cadd(ephi,field+1,ephi);
} /*calc_ptrn*/

/*
 * Calculation apprximation factor
 *
 */
calc_apprx_factor(angle,lpara,nlevels,apprx_level,iground)
     Angles *angle;
     LevelParameters *lpara;
     int nlevels;
     int apprx_level;
     int iground;
{
  int ilev;
  Rtyp phase1, phase2; /* phases for direct and secondary rays */
  Rtyp t1,t2,t3,t4,T_phi,T_theta;
  complex c0,c1,c2;
  complex *pf;
  LevelParameters *ptrl, *pptrl;

  switch (apprx_level) {
  case PTN_APPRX_DIPOLE:  /* using far field Green's function */
    pf=lpara->factor;
    t1=sqrt(fabs(lpara->beta_r-angle->sin_t*angle->sin_t));
    phase1=lpara->kz*t1;
    Cexpj(phase1,&c1);
    if (fabs(angle->cos_t) < EPS) { /* on the horizen */
      pf[0].Re=pf[0].Im=0.0;
      pf[1].Re=pf[1].Im=0.0;
    }
    else  {
      T_theta= 1.0/lpara->beta_r;
      t2= lpara->constant*angle->cos_t*c1.Im/(angle->cos_t*angle->cos_t*c1.Im*c1.Im+t1*t1*c1.Re*c1.Re);
      pf[1].Re= t1*c1.Re*t2;
      pf[1].Im= angle->cos_t*c1.Im*t2;
      t2= T_theta*t1/(angle->cos_t*angle->cos_t*c1.Re*c1.Re + t1*t1*c1.Im*c1.Im);
      pf[0].Re = (c1.Re*c1.Re+lpara->beta_r*c1.Im*c1.Im)*t1*angle->cos_t*t2;
      pf[0].Im = (t1*t1-lpara->beta_r*angle->cos_t*angle->cos_t)*c1.Re*c1.Im*t2;
      Cmult(pf, pf+1, pf);
      pf[0].Im *= -1.0;
      pf[1].Im *= -1.0; 
    }
    break;    	
  case PTN_APPRX_NO_DIELECTRICS:	
    for (ilev=0; ilev < nlevels; ilev++) {
      ptrl=lpara+ilev;
      pf=ptrl->factor;
      if (iground == YES) {
        pf[0].Re=pf[1].Re= 0.0;
        pf[1].Im= ptrl->constant*sin(ptrl->kz*angle->cos_t);
	pf[0].Im= pf[1].Im*angle->cos_t;
      }
      else {
        pf[1].Re= ptrl->constant*sin(ptrl->kz*angle->cos_t);
        pf[1].Im= -ptrl->constant*cos(ptrl->kz*angle->cos_t);
	pf[0].Re= pf[1].Re*angle->cos_t;
	pf[0].Im= pf[1].Im*angle->cos_t;
      }
    }
    break;
  case PTN_APPRX_RECIPROCITY:
  default:
    for (ilev=0,ptrl=lpara;
	 ilev < nlevels; 
	 ilev++, ptrl++) {
      /* find the angles in each layer */
      ptrl->sin_t= ptrl->eta_r * angle->sin_t;
      if (ptrl->sin_t > 1.0) 	ptrl->total_reflection= YES;
      else {
	ptrl->total_reflection= NO;
	ptrl->cos_t= sqrt((1.0+ptrl->sin_t)*(1.0-ptrl->sin_t));
	/* find the reflection and transmission coefficients */
	if (ilev == 0) {
	  t1= ptrl->eta_r * ptrl->cos_t;
	  t2= angle->cos_t;
	  t3= ptrl->eta_r * angle->cos_t;
	  t4= ptrl->cos_t;
	}
	else {
	  t1= ptrl->eta_r * ptrl->cos_t;
	  t2= (ptrl-1)->eta_r * (ptrl-1)->cos_t;
	  t3= ptrl->eta_r * (ptrl-1)->cos_t;
	  t4= (ptrl-1)->eta_r * ptrl->cos_t;
	}	
	ptrl->tt_01= 2.0*t1/(t1+t2);
	ptrl->rt_10= (t2-t1)/(t1+t2);
	ptrl->tp_01= 2.0*t3/(t3+t4);
	ptrl->rp_10= (t4-t3)/(t3+t4);
      } /* else */
    } /* for */

    for (ilev=0,ptrl=lpara;
	 ilev < nlevels && ptrl->total_reflection == NO; 
	 ilev++, ptrl++) {
      if (ilev == nlevels-1) {
	if (iground == YES) {
	  ptrl->rt_12= -1.0;
	  ptrl->rp_12=  1.0;
	}
	else {
	  ptrl->rt_12= -(lpara->rt_10);
	  ptrl->rp_12= -(lpara->rp_10);
	}
      }
      else {
	ptrl->rt_12= - (ptrl+1)->rt_10;
	ptrl->rp_12= - (ptrl+1)->rp_10;
      }
      Cexpj(-2.0*ptrl->h_r*ptrl->cos_t,&c1);
      /* theta component */
      t3= ptrl->rt_10*ptrl->rt_12;
      t1= 1.0 - t3*c1.Re;
      t2= -t3*c1.Im;
      t3= angle->cos_t*ptrl->constant*ptrl->tt_01/(t1*t1+t2*t2);
      ptrl->factor[0].Re=  t3*t1;
      ptrl->factor[0].Im= -t3*t2;
      /* phi component */
      t3= ptrl->rp_10*ptrl->rp_12;
      t1= 1.0 - t3*c1.Re;
      t2= -t3*c1.Im;
      t3= ptrl->constant*ptrl->tp_01/(t1*t1+t2*t2);
      ptrl->factor[1].Re=  t3*t1;
      ptrl->factor[1].Im= -t3*t2;
      if (ilev > 0) {
	pptrl= ptrl-1;
	Cexpj(-pptrl->h_r*pptrl->cos_t, &c1);
	Cmult( pptrl->factor,&c1,&c2);
	Cmult( ptrl->factor, &c2, ptrl->factor);
	Cmult( pptrl->factor+1,&c1,&c2);
	Cmult( ptrl->factor+1, &c2, ptrl->factor+1);
      } /* if */
    } /* for */
    break;
  } /* switch */
} /* calc_apprx_factor */

calc_ray_sum(c1,c2,n)
complex *c1;
complex *c2;
int n;
{
  int i=1;
  complex ai,ai_1;

  if (n < 1) {
    c2->Re=0.0;
    c2->Im=0.0;
    return;
  }
  c2->Re=1.0;
  c2->Im=0.0;
  ai_1.Re=1.0;
  ai_1.Im=0.0;
  while (i < n) {
    Cmult(c1,&ai_1,&ai);
    Cadd(&ai,c2,c2);
    ai_1=ai;
    i++;
  }
}
change_jxy_2_rect(pd0,istep)
     Pdat *pd0;
     int istep;
/****************************************************************/
/* Change current from polar values to rect values              */
/*                                                              */
/* (C) COPYRIGHT 1992 SONNET SOFTWARE, INC. ALL RIGHTS RESERVED.*/
/****************************************************************/
/* tw20apr93 */
{ 
  int ix,iy,dat_head;
  float *dmxp,*dpxp,*dmyp,*dpyp,*dmzp,*dpzp;
  complex ct1,ct2,ct3;
  complex Jxi,Jyi,Mzi;
  Pdat *pd;

  pd=pd0;
  while(pd) {
    dat_head = pd->nx*pd->ny*istep;
    dmxp=pd->jx_mg+dat_head; dpxp=pd->jx_ph+dat_head;
    dmyp=pd->jy_mg+dat_head; dpyp=pd->jy_ph+dat_head;
    dmzp=pd->ez_mg+dat_head; dpzp=pd->ez_ph+dat_head;
    for (iy=0; iy<pd->ny; iy++) {
      for (ix=0; ix<pd->nx; ix++) {
	Jxi.Re= *dmxp; Jxi.Im= *dpxp;
	Jyi.Re= *dmyp; Jyi.Im= *dpyp; 	
	Mzi.Re= *dmzp; Mzi.Im= *dpzp; 	
/*	if (Mzi.Re < *(pd0->stats+EZ_MG_SIG)) Mzi.Re=0.0; */
	if (iy > 0 && iy < pd->ny-1 &&  
	    ix > 0 && ix < pd->nx-1) Mzi.Re=0.0;  
	/* convert complex numbers in rect format from polar format  */
	Cpolar2rect(&Jxi,&ct1);
	Cpolar2rect(&Jyi,&ct2);
	Cpolar2rect(&Mzi,&ct3);
	*dmxp= ct1.Re; *dpxp= ct1.Im;
	*dmyp= ct2.Re; *dpyp= ct2.Im;
	*dmzp= ct3.Re; *dpzp= ct3.Im;
	dmxp++; dpxp++; dmyp++; dpyp++; dmzp++; dpzp++;
      }/*iy*/
    }/*ix*/
    pd=pd->next;
  }/*while*/
} /*change_jxy_2_rect*/

calc_level_parameters(kfact,jd,lpara)
     double kfact;
     Jdat *jd;
     LevelParameters *lpara;
{
  LevelParameters *ptrl;
  int ilev,nlevels,i;
  double kz,t1;
  double eps_r, mu_r;
  complex c0,c1;

  c0.Re=0.0; c0.Im=0.0;
  nlevels=jd->box->Num_levels;
  ptrl=lpara+nlevels-1;
  kz=0.0;
  for (ilev=nlevels-1; ilev >= 0; ilev--) {
    kz += *(jd->box->H+ilev+1)*kfact;
    ptrl->kz = kz;
    ptrl--;
  }
  
  for (ilev=0,  ptrl=lpara; ilev < nlevels; ilev++,ptrl++) {
    eps_r= *(jd->box->Eps+ilev+1)/(*(jd->box->Eps));
    mu_r= *(jd->box->Mu+ilev+1)/(*(jd->box->Mu));
    ptrl->eta_r= sqrt(mu_r/eps_r);
    ptrl->beta_r= sqrt(mu_r*eps_r);
    ptrl->h_r=*(jd->box->H+ilev+1)*kfact*ptrl->beta_r;
    ptrl->constant= 60.0/kfact;
    ptrl->total_reflection=NO;
    for (i=0; i < 2; i++) ptrl->factor[i]=c0;
  }
} /* calc_level_parameters */

int change_ptn_coord(angle,iground)
/*****************************************************************************/
/* Change current coordinates to spherical coordinates                       */
/* (C) COPYRIGHT 1992 SONNET SOFTWARE, INC, ALL RIGHTS RESEARVED             */
/****************************************************************************/
/* tw06may92 */
Angles *angle;
int iground;
{
  Rtyp tt;

  switch(angle->coord) {
  case PTN_COORD_SPHERICAL:
    if (iground == YES && angle->theta > (PI_OVER_2+0.01)) return (NO);
    break;
  case PTN_COORD_EARTH:
    if ((angle->phi+PI_OVER_2) < -1.0e-6) {
      if (angle->theta < 0) 
	angle->theta=-angle->theta;
      else
	angle->phi = -angle->phi;
    }
    else {
      if (angle->theta < 0) {
	angle->theta=-angle->theta;  
	if (angle->phi < 0) angle->phi +=  PI;
	else                angle->phi -=  PI;
      }
    }
    break;
  case PTN_COORD_POLAR:
    tt=sqrt(fabs(angle->theta*angle->theta+angle->phi*angle->phi));
    if (tt > PI_OVER_2)  return(NO);
    else {
      if (tt < TINY) {
	angle->theta=0.0;
	angle->phi=0.0;
      }
      else           {
	angle->phi=atan2(angle->phi,angle->theta); 
	angle->theta=tt;
      }
    }
    break;
  case PTN_COORD_ARRAY:
    tt=angle->theta*angle->theta+angle->phi*angle->phi;
    if (tt > 1.0) return (NO);
    else {
      angle->sin_t=sqrt(fabs(tt)); angle->cos_t=sqrt(fabs(1.0-tt));
      if (angle->sin_t < TINY) {
	angle->sin_p=0.0; angle->cos_p=1.0;
      }
      else {
	angle->sin_p=angle->phi/angle->sin_t;
	angle->cos_p=angle->theta/angle->sin_t;
      }
      return(0);
    }
  case PTN_COORD_ARRAY_SINE:
    angle->theta=sin(angle->theta); angle->phi=sin(angle->phi);
    tt=angle->theta*angle->theta+angle->phi*angle->phi;
    if (tt > 1.0) return (NO);
    else {
      angle->sin_t=sqrt(fabs(tt)); angle->cos_t=sqrt(fabs(1.0-tt));
      if (angle->sin_t < TINY) {
	angle->sin_p=0.0; angle->cos_p=1.0;
      }
      else {
	angle->sin_p=angle->phi/angle->sin_t;
	angle->cos_p=angle->theta/angle->sin_t;
      }
      return(0);
    }
  }
  /* when theta = 0, phi is undefined, set it to be 0 */
/*   if (fabs(angle->theta) < TINY) angle->phi=0.0; */

  angle->sin_t=sin(angle->theta); angle->cos_t=cos(angle->theta);
  angle->sin_p=sin(angle->phi);   angle->cos_p=cos(angle->phi);
  return(0);
}
calc_input_power(jd,ifreq,power)
Jdat *jd;
int ifreq;
float *power;
/****************************************************************/
/* Calculate pattern given current distribution and direction   */
/* Exective Routine                                             */
/* (C) COPYRIGHT 1992 SONNET SOFTWARE, INC. ALL RIGHTS RESERVED.*/
/****************************************************************/
/* tw15feb93 */
{
  int iport;
  double vm,omega,t1,t2,Zr,Zi,pwr;
  Mat_ptr Z;
  extern Mat_ptr Mat_open();
  extern double Mat_get_re();
  extern double Mat_get_im();
  
  Z=Mat_open(jd->nport,jd->nport);
  Mat_move((jd->fdat+ifreq)->cans,Z);
  stoz(Z);
  omega=TWO_PI* (jd->fdat+ifreq)->freq*1e+06;
  pwr=0.0;
  for (iport=0; iport<jd->nport; iport++) {
    vm=jd->vsrc_mg[iport];
    Zr=Mat_get_re(iport,iport,Z);
    Zi=Mat_get_im(iport,iport,Z);
    calc_port_cir(omega,jd->rsrc[iport],jd->xsrc[iport],
		  jd->lsrc[iport],jd->csrc[iport],&t1,&t2);
    t1 += Zr;
    t2 += Zi;
    pwr += Zr*vm*vm/(t1*t1+t2*t2);
  }
  Mat_close(Z);
/*  t1=2.0*omega/VEL_LIGHT;
  pwr = sqrt(fabs(pwr*TWO_PI/PTN_IMPEDANCE))*t1; */
  pwr = sqrt(fabs(60.0*pwr));
  *power = (float)pwr;
}

calc_port_cir(omega,r,x,l,c,zr,zi)
double omega,r,x,l,c;
double *zr,*zi;
{
  double t0,t1,t2,tt;

  t0=x+omega*l;
  t1=1.0-omega*c*t0;
  t2=omega*c*r;
  tt=t1*t1+t2*t2;
  *zr=(r*t1+t0*t2)/tt;
  *zi=(t0*t1-r*t2)/tt;
}

