
/* Copyright (c) 1990 Sonnet Software, Inc., All rights reserved. */
/******************************************************************/

/* %W% last mod on %G% at %U% */
/* Full SCCS path name: (%P%) */

/*******************************************************************

	NAME: rkeyfile.i
	
	AUTHOR: K. Schroeder
	
	PURPOSE: Handles read file manipulations.
	
	CONTENTS:
		read_key_file() - read in the keyfile
		get_key_info() - Get the key information.

 	INTERNAL:
		bad_keyfile() - Tell's user error encounted in keyfile
		load_key_values() - Set value that is inputed from a 
				    keyfile line
		init_prog() - Initialize program key values.
		getline() - get a single line from a file
		cut() - Grabs out a specific field a string 
		get_license_id() _ get license identifier.
		get_sn() _ get serial numbers
	
*******************************************************************/

#include <errno.h>

/* definitions  */
#define TAB     '\t'

/* external	*/

/* structures and global/semiglobal variable */

LOCAL_FN int
read_key_file(
#if NeedFunctionPrototypes
	Key *key,
	int print_error
#endif
);

LOCAL_FN void
init_prog(
#if NeedFunctionPrototypes
	Prog *prog,
	char *name
#endif
);

LOCAL_FN int
load_key_values(
#if NeedFunctionPrototypes
	Prog *prog,
	char *cmd,
	char *value,
	char *filename,
	char *line,
	int linenu
#endif
);

LOCAL_FN void
bad_keyfile(
#if NeedFunctionPrototypes
	char *error_msg,
	char *filename,
	char *line,
	int linenu
#endif
);

LOCAL_FN int 
getline(
#if NeedFunctionPrototypes
	FILE *file,
	char *filename,
	int linenu,
	char *line,
	int maxlen
#endif
);

LOCAL_FN void
cut(
#if NeedFunctionPrototypes
	char *instr,
	int field_nu,
	char *outstr,
	int maxlen
#endif
);

/*******************************************************************
	NAME: get_key_info
	
	DESCRIPTION: 
		Get the key information.

		Print out error messages if keyfile either does not 
		exists or is non-readable only if print_error = 
		PRINT_ERROR.
		Else do not print out an error message
		and initialize key and oldkey structures.

		If read_file is TRUE initialize key from keyfile.
		otherwise set values assuming no options were
		selected.

	RETURNS: TRUE - if read file in properly
		 FALSE - if either could not read in file and 
			 print_error = PRINT_ERROR or
			 if file has an improper format

*******************************************************************/


int
get_key_info(key,print_error,read_file)
Key *key;
int print_error;
int read_file;
{
	int rc;

	rc = TRUE;
	strcpy(key->key,"UNKNOWN");
	strcpy(key->release,"0.0" );
	strcpy(key->license,"UNKNOWN");
	key->nu_sn = NONE;
	key->sn = NULL;

	init_prog(&key->em,"Em");
	init_prog(&key->xgeom,"Xgeom");
	init_prog(&key->emvu,"Emvu");
	init_prog(&key->patvu,"Patvu");

	if ( read_file ) {
		rc = read_key_file(key,print_error);
	} else {
		key->file = FALSE;
	}
	return (rc);
}

/*******************************************************************
	NAME: read_key_file
	
	DESCRIPTION: 
		Read in the keyfile, then put values into oldkey.
		Print out error messages if keyfile either does not 
		exists or is non-readable only if print_error = PRINT_ERROR.
		Else do not print out an error message.

	RETURNS: TRUE - if read file in properly
		 FALSE - if either could not read in file and 
			 print_error = PRINT_ERROR or
			 if file has an improper format


*******************************************************************/

#define CASE(b) else if ( (strcmp(string,b) == 0) )
#define SWITCH(a) strcpy(string,a); if (FALSE) 
#define LINE_LEN	256

LOCAL_FN int
read_key_file(key,print_error)
Key *key;
int print_error;
{
	FILE *file;
	int linenu;
	int rc;
	int i,len;
	char c;
	char prog[QUESTION_LEN];
	char cmd[QUESTION_LEN];
	char value[QUESTION_LEN];
	char line[LINE_LEN];
	char string[QUESTION_LEN];

	/* open key file */
	file = fopen(key->filename,"r"); 
	if ( file == NULL ) { 
		if (print_error == PRINT_ERROR ) {
			if ( errno == ENOENT ) {
				fprintf(stderr,"The Sonnet Software has not yet been unlocked.  To unlock the \n");
				fprintf(stderr,"software, contact your Sonnet Software representative.\n");
				return FALSE;
			} else {
				fprintf(stderr,"Could not read file %s\n",key->filename);
				perror( AUX_ERR_MSG );
				return FALSE;
			}
		}
		key->file = FALSE; 
		return TRUE;
	} else {
		key->file = TRUE;
	}

	linenu = 1;
	
	while ( getline(file,key->filename,linenu,line,LINE_LEN) != EOF ) {
		cut(line,1,prog,QUESTION_LEN);
		cut(line,2,cmd,QUESTION_LEN);
		cut(line,3,value,QUESTION_LEN);
		
		SWITCH ( prog );
		CASE( key->em.name ) {
			rc = load_key_values(&key->em,cmd,value,key->filename,line,linenu);
			if ( rc == FALSE ) {
				fclose(file);
				return FALSE;
			}
		}
		CASE( key->xgeom.name ) {
			rc = load_key_values(&key->xgeom,cmd,value,key->filename,line,linenu);
			if ( rc == FALSE ) {
				fclose(file);
				return FALSE;
			}
		}
		CASE( key->emvu.name ) {
			rc = load_key_values(&key->emvu,cmd,value,key->filename,line,linenu);
			if ( rc == FALSE ) {
				fclose(file);
				return FALSE;
			}
		}
		CASE( key->patvu.name ) {
			rc = load_key_values(&key->patvu,cmd,value,key->filename,line,linenu);
			if ( rc == FALSE ) {
				fclose(file);
				return FALSE;
			}
		}
		CASE( "key" ) {
			strncpy(key->key,value,KEY_LENGTH-1);
		}
		CASE( "license" ) {
			strncpy(key->license,value,LICENSE_LEN-1);
			if ( strcmp(key->license,"") == 0 ) {
				strcpy(key->license,"UNKNOWN");
			} else {
				len = strlen(key->license);
				for (i = 0; i< len; ++i ) {
					c = key->license[i];
					if ( ( ! isalnum(c))  && c != '.' ) {
						bad_keyfile("Invalid license id",
							    key->filename,line,linenu);
						if ( rc == FALSE ) {
							fclose(file);
							return FALSE;
						}
					}
				}
			}
		}
		CASE( "sn" ) {
			add_sn(key,value);
		}
		CASE( "release" ) {
			strncpy(key->release,value,RELEASE_LEN);
			if ( strcmp(key->release,"") == 0 ) {
				bad_keyfile("Invalid release number",
					    key->filename,line,linenu);
			}
		}
		CASE( "#" ) {
			/* SKIP COMMENT */
		}
		CASE( "# " ) {
			/* SKIP COMMENT */
		} else {
			bad_keyfile("Unreconized first field",
				    key->filename,line,linenu);
			fclose(file);
			return FALSE;
		}
		++linenu;
	}
	fclose(file);
	return TRUE;
}

/*******************************************************************
	NAME: init_prog
	
	DESCRIPTION: 
		Initialize the Key values for a specific program.
	
	RETURNS: none

*******************************************************************/

LOCAL_FN void
init_prog( prog,name )
Prog *prog;
char *name;
{
	prog->have_prog  = FALSE;
	prog->alike  = ALIKE_UNIQ;
	prog->expir_date = FALSE;
	prog->nu_nodes   = NONE;
	prog->nodes      = NULL;
	prog->subs       = NONE;
	prog->name       = name;
}

/*******************************************************************
	NAME: load_key_values
	
	DESCRIPTION: 
		A line was read in from keyfile.  The program determines
		the prog value.  The command and associated value is
		stored in cmd, value.
	
	RETURNS: TRUE - if succeeded in converting the line.
		 FALSE - if unsuccessful in converting the line.

*******************************************************************/

LOCAL_FN int
load_key_values(prog,cmd,value,filename,line,linenu)
Prog *prog;
char *cmd;
char *value;
char *filename;	/* file name of keyfile */
char *line;	/* single line from keyfile. */
int linenu;	/* line number in keyfile    */
{
	char string[QUESTION_LEN];
	int digits;
	int rc;

	prog->have_prog = TRUE;

	SWITCH ( cmd );
	CASE( "lock" ) {
		if ( strcmp(value,"em") == 0 ) {
			prog->alike = ALIKE_EM;
		} else {
			bad_keyfile("Bad lock value",
				    filename,line,linenu);
			return FALSE;
		}
	}
	CASE( "date" ) {
		prog->expir_date = TRUE;
		rc = get_date(&prog->expir,value);
		if ( rc == FALSE ) {
			bad_keyfile("Bad expiration date",
				    filename,line,linenu);
			return FALSE;
		}
	}
	CASE( "node") {
		add_node(prog,value);
	}
	CASE( "subs") {
		rc = getnumber(&prog->subs,&digits,value);
		if ( rc == FALSE ) {
			bad_keyfile("Bad value for number of subsections",
				     filename, line,linenu);
			return FALSE;
		}
	} else {
		bad_keyfile("Unreconized second field value",
			     filename, line,linenu);
	}
	return TRUE;
}

/*******************************************************************
	NAME: bad_keyfile
	
	DESCRIPTION: 
		Tell the user that we encounted an error in the 
		keyfile.
	
	RETURNS: NONE

*******************************************************************/

LOCAL_FN void
bad_keyfile(error_msg,filename,line,linenu)
char *error_msg; /* error message to be displayed to user */
char *filename;	/* name of the keyfile			*/
char *line;	/* Line error was encounted on		*/
int linenu;	/* line number error was encounted on	*/
{
	fprintf(stderr,"Encounted an error in '%s' on line %d\n",filename,linenu);
	fprintf(stderr,"%s\n",error_msg);
	fprintf(stderr,"Error occured in the following line:\n");
	fprintf(stderr,"%s\n",line);
}


/*******************************************************************
	NAME: getline
	
	DESCRIPTION: 
		Reads in a single line from file filename.
		If line lenth is exceeded, then an error
		message is printed out.
	
	RETURNS: 
		EOF - if end of file reached
		0 - if end of file NOT reached.

*******************************************************************/

LOCAL_FN int 
getline(file,filename,linenu,line,maxlen)
FILE *file;		/* file descripter for open file to be read */
char *filename;		/* file name */
int linenu;		/* line number being read - used in error msg */
char *line;		/* RETURN: line read in */
int maxlen;		/* Maximum length of line */
{
	int i;
	int ch;
	
	i=0;
	while (TRUE) {
		ch=getc(file);
		if (i >= maxlen) {
			fprintf(stderr,"ERROR: Line %d in file %s exceeded maximum line length\n",
				linenu,filename);
			fprintf(stderr,"Maximum line length is: %d\n",maxlen);
			line[i-1]=0;
			fprintf(stderr,"Beginning part of this line is:\n");
			fprintf(stderr,"%s\n",line);
			exit (1);
		}
		switch (ch) {
			case EOF:	return(EOF);
			case '\n':      line[i]=0;
					return(ch);
			default:        line[i++] = (char)ch;
		}
	}
}

/*******************************************************************
	NAME: cut
	
	DESCRIPTION: 
		Grabs out a specific field in the instr
		string and copies result to outstr.
		Field are deliminated by TAB.  If a field is
		not found then instr is set to NULL.  There
		may not be more than maxlen characters in
		the field.  If there are, then the field is
		truncated when copied.
	
	RETURNS: 

*******************************************************************/

LOCAL_FN void
cut(instr,field_nu,outstr,maxlen)
char *instr;
int field_nu;
char *outstr;
int maxlen;
{
	int field_len;
	char *field;
	
	/* skip over the fields */
	while ( field_nu != 1 ) {
		if ( *instr == TAB ) {
			--field_nu;
		} else {
			if ( *instr == '\0' ) {
				*outstr = '\0';
				return;
			}
		}
		++instr;
	}
	field = instr;
	field_len = 0;
	while ( (*instr != TAB) && 
	        (*instr != '\0') && 
		(field_len <= maxlen) ) {
		
		++field_len;
		++instr;
	}
	
	(void) strncpy(outstr,field,field_len);
	outstr[field_len] = '\0';
}

/*******************************************************************
	NAME: get_keyfile_name
	
	DESCRIPTION: Get the keyfile name.

		1. check if SONNET_DIR variable is set if so
		   set keyfile name to
			$SONNET_DIR/<KEYFILE_NAME>

	RETURNS: 
		TRUE - was able to get key file name
		FALSE - was not able to get key file name, because
			SONNET_DIR not defined

*******************************************************************/


int
get_keyfile_name(filename)
char *filename;
{
	char *sonnet_dir;
	extern char *getenv();

/* ---------- RELEASE 2.1d BEGIN --------- */
	sonnet_dir = get_sonnet_dir(TRUE);
	if ( sonnet_dir == NULL ) {
		return (FALSE);
	}
/* ---------- RELEASE 2.1d END --------- */

	strncpy(filename,sonnet_dir,FILE_LEN-16);
	strcat(filename,"/");
	strcat(filename,DEFAULT_KEYFILE_NAME);
	return (TRUE);
}

/* ------------------ RELEASE 2.1d BEGIN -------------------- */



/*******************************************************************

	Return user's license id 

*******************************************************************/
char *
get_license_id()
{
	Key key;
	int rc;
	char *license_id;

	rc = get_keyfile_name(key.filename);
	if ( rc == FALSE ) {
		return( KEY_NOT_DEFINED	);
	}

	rc = get_key_info(&key,DO_NOT_PRINT_ERROR, TRUE);
	if ( rc == FALSE ) {
		return( KEY_NOT_DEFINED	);
	}

	/* Save license ids */
	license_id = get_license_id_string();
	strcpy(license_id,key.license);
	return (license_id);
}

/*******************************************************************

	Return user's serial numbers

*******************************************************************/
void
print_sn(file)
FILE *file;
{
	Key key;
	int rc;
	char *license_id;
	int i;

	rc = get_keyfile_name(key.filename);
	if ( rc == FALSE ) {
		fprintf(file,"SN:/tnone\n");
		return;
	}

	rc = get_key_info(&key,DO_NOT_PRINT_ERROR, TRUE);
	if ( rc == FALSE ) {
		fprintf(file,"SN:/tnone\n");
		return;
	}

	/* Save license ids */
	for (i=0; i < key.nu_sn; ++i ) {
		fprintf(file,"SN:/t%s\n",key.sn[i]);
	}
}

char *
get_license_id_string()
{
	static char license_id[LICENSE_LEN];
	return( license_id );
}
