 

/*******************************************************************
	NAME: security()
	
	DESCRIPTION: 
		Check if user is allow to execute the program or not.
		If not, an error messages is printed stating why
		the user is not authorized to run the program.
		If print_time is TRUE and the software is time locked,
		then print out the expiration date to stderr.


	Reasons for Authorization to fail:
		- Key file does not exists 
		- Sonnet directory does not exists
		- could not get key info.
		- invalid password.
		- Not authorized on node
		- time limit expired
		- release not supported 
	
	
	RETURNS: 
		TRUE - user is authorized to run the program.
		       User may still be limited in program use.
		FALSE - user is not authorized to run the program.

*******************************************************************/


/* check timeout and protection */
#include "keyfile.h"
#include <ctype.h>
#include <sys/types.h>
#include <errno.h>

#ifdef DOS
#include "time.h"
#else
#endif

/* Time.h is defined on Sun, IBM6000 - use to be <sys/time.h> */
#include <time.h>

#include "son.h"
#include "unlcki_p.h"

#define KEY_NOT_DEFINED	"Not Defined"

static int  em_subs;

int
security(prog_id,rel_num,print_time,keyfile)
int prog_id;		/* Indicates program to test			*/
			/* =1 for EM					*/
			/* =2 for XGEOM					*/
			/* =3 for EMVU					*/
			/* =4 for PATVU/PATGEN				*/
char *rel_num;		/* release number id of program 		*/
int print_time;		/* If TRUE, print out expiration date		*/
char *keyfile;		/* Name of keyfile, if NULL then use default	*/
{
	Prog *prog;
	int rc;
	int found_node;
	int i;
	char *node_id;
	float prog_num, key_num;
	Key key;
	char *sonnet_dir;
	char filename[FILE_LEN];
	char *license_id;

	em_subs = -1;
	license_id = get_license_id_string();
	strcpy(license_id,KEY_NOT_DEFINED);

	if ( keyfile == NULL ) {
		rc = get_keyfile_name(key.filename);
		if ( rc == FALSE ) {
			return(FALSE);
		}
	} else {
		strcpy(key.filename, keyfile);
	}

	sonnet_dir=get_sonnet_dir(TRUE);
	if ( sonnet_dir == NULL ) {
		return (FALSE);
	}

	/*
	 *	Verify that SONNET_DIR points to the sonnet directory
	 */

	strncpy(filename,sonnet_dir,FILE_LEN-16);
	strcat(filename,"/sonnet.ver");
	if ( access(filename,0) == -1 ) {
		fprintf(stderr,"ERROR: The environment variable SONNET_DIR does not point to the sonnet\n");
		fprintf(stderr,"        directory.  See the Sonnet installation manual on how to set\n");
		fprintf(stderr,"        the SONNET_DIR variable.\n");
		return(FALSE);
	}

	/*
	 *	Verify that data/installem.flg does not exists
	 */
	strncpy(filename,sonnet_dir,FILE_LEN-16);
	strcat(filename,"/data/installem.flg");
	if ( access(filename,0) == 0 ) {
		fprintf(stderr,"\n");
		fprintf(stderr,"ERROR:   The Sonnet Software installation has not been completed.\n");
		fprintf(stderr,"	 You should have your administrator run the 'installem' command.\n");
		fprintf(stderr,"\n");
		return (FALSE);
	}

	rc = get_key_info(&key,PRINT_ERROR, TRUE);
	if ( rc == FALSE ) {
		return (FALSE);
	}

	switch (prog_id) {
	case PROG_EM:
		prog = &key.em;
		break;

	case PROG_XGEOM:
		if ( key.xgeom.alike == ALIKE_EM ) {
			prog = &key.em;
		} else {
			prog = &key.xgeom;
		}
		break;

	case PROG_EMVU:
		if ( key.emvu.alike == ALIKE_EM ) {
			prog = &key.em;
		} else {
			prog = &key.emvu;
		}
		break;

	case PROG_PATVU:
		if ( key.patvu.alike == ALIKE_EM ) {
			prog = &key.em;
		} else {
			prog = &key.patvu;
		}
		break;
	}

	/*
	** Set method to encode the password.
	*/
	sscanf(key.release,"%f",&key_num);
	if ( key_num >= 2.2 ) {
		rc = check_key2(&key);
	} else {
		rc = check_key(&key);
	}

	if ( rc != KEY_GOOD) {
		fprintf(stderr,"%s: You have an invalid password.\n",prog_name);
		sleep (2);
		return (FALSE);
	}

	if ( prog->expir_date == FALSE && prog->nu_nodes  == NONE ) {
		fprintf(stderr,"You are not authorized to run %s.\n",prog_name);
		return (FALSE);
	}

	/* print out expiration date */
	if ( prog->expir_date == TRUE && print_time ) {
		display_expiration_date(prog);
	}

	if ( prog->expir_date ) {
		rc = check_date(&prog->expir);
		if ( rc != TRUE) {
			return (FALSE);
		}
	}

	if ( prog->nu_nodes != NONE ) {
		node_id = get_node_id();
		if ( node_id == NULL ) {
			fprintf(stderr,"%s: Could not get the node if for this computer\n",prog_name);
			exit (1);
			
		}
		tolowerstr(node_id);
		found_node = FALSE;
		for (i = 0; i < prog->nu_nodes; ++i ) {
			tolowerstr(prog->nodes[i]);
			if ( strcmp(node_id,prog->nodes[i]) == 0 ) {
				found_node = TRUE;
				break;
			}
		}

		if ( ! found_node ) {
			fprintf(stderr,"%s: Not authorized on this node\n",prog_name);
			fprintf(stderr,"%s: Node id for this computer is: %s\n",prog_name,node_id);
			return (FALSE);
		}
	}

	/* check release number */
	sscanf(rel_num,"%f",&prog_num);
	sscanf(key.release,"%f",&key_num);
	if ( prog_num > key_num ) {
		fprintf(stderr,"%s: Release number for program is not supported by your key.\n",prog_name);
		fprintf(stderr,"Contact a Sonnet Software repesentative for a new key.\n");
		return (FALSE);
		
	}

	/* determine the limit of the number of subsection for Em */
	if ( key.em.subs == NONE ) {
		em_subs = 0;
	} else {
		em_subs = key.em.subs;
	}

	/* Save license ids */
	strcpy(license_id,key.license);

	return (TRUE);
}

void
display_expiration_date(prog)
Prog *prog;
{
	time_t timeofday;
	struct tm *tm;
	int expirs,today;
	static char *month_str[] = {
                "January","February","March","April","May","June",
                "July","August","September","October","November","December"
        };

	/* Get days in year */
	timeofday = time( (time_t *)0 );
	tm = localtime( &timeofday );
	today = tm->tm_mday 
			+ 30  * tm->tm_mon 
			+ 365 * tm->tm_year;
	expirs = prog->expir.day
			+ 30  * prog->expir.month
			+ 365 * (prog->expir.year - 1900);


	/*************************************************************/
	/*							     */
	/* NOTE: THE TIME OUT ERROR MESSAGE ARE REFERENCED IN testem */
	/*							     */
	/*************************************************************/

	/* Print out expiration date if within a month only */
	if ( expirs < today ) {
		fprintf(stderr,"ERROR: program expired on %s %d, %d.\n",
			month_str[prog->expir.month],
			prog->expir.day,
			prog->expir.year);
	}
	else if ( expirs - today < 7 ) {
		fprintf(stderr,"WARNING: This program expires within a week on %s %d, %d.\n",
			month_str[prog->expir.month],
			prog->expir.day,
			prog->expir.year);
	}

	/* Print out expiration date if within a month only */
	else if ( expirs - today < 30 ) {
		fprintf(stderr,"This program expires on %s %d, %d.\n",
			month_str[prog->expir.month],
			prog->expir.day,
			prog->expir.year);
	}

}

/* Return the limit of the number of subsections in the .key file.
   If no limit exists returns 0.  If keyfile is invalid or does not
   exists return -2.
*/
int get_sub_limit()
{
        return( em_subs );
}

#ifndef EXCLUDE_SECURITY_BASE_FILES

#include "rkeyfile.i"
#include "date.i"
#include "crypt.i"
#include "crypt2.i"
#include "unnode.i"
#include "usn.i"
#include "wkeystr.i"
#include "getnode.i"

#endif /* EXCLUDE_SECURITY_BASE_FILES */
